import json
import warnings

import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error, explained_variance_score, mean_squared_error, median_absolute_error, \
    r2_score, mean_absolute_percentage_error
from sklearn.model_selection import KFold
from sklearn.preprocessing import StandardScaler
from sklearn.svm import LinearSVR

warnings.filterwarnings('ignore')


class Evaluation:
    def __init__(self, pro_name='COAD', model_name='SVAE'):
        self.pro_name = pro_name
        self.model_name = model_name
        self.final_days_path = 'dataset/raw_data/final_days/'
        self.fc_data_path = 'dataset/fc_data/'
        self.predict_result_path = 'dataset/fc_data/predict/'
        self.detail_clinical_path = 'dataset/detail/'

    def predict(self, i):
        np.set_printoptions(formatter={'float': '{: 0.2f}'.format})
        if i == 0:
            matrix = pd.read_table(self.fc_data_path + self.pro_name + '.txt', header=None,
                                   sep='\t').values
        else:
            matrix = pd.read_table(self.fc_data_path + self.model_name + '_' + self.pro_name + '_' + str(i) + '.txt',
                                   header=None, sep='\t').values
        days = pd.read_table(self.final_days_path + self.pro_name + '.txt', header=None, sep='\t').values
        m = StandardScaler()
        days = m.fit_transform(days)
        matrix = matrix[:, :-1]
        days = days.squeeze()
        kf = KFold(n_splits=10, random_state=0, shuffle=True)
        print('[KFold_random_seed]: {}'.format(kf.random_state))
        predict_days = np.zeros((len(days)))
        for train, val in kf.split(matrix, days):
            t_data, t_label = np.array(matrix)[train], np.array(days)[train]
            v_data, v_label = np.array(matrix)[val], np.array(days)[val]
            lsvr = LinearSVR()
            lsvr.fit(t_data, t_label)
            predict = lsvr.predict(v_data)
            predict = m.inverse_transform(predict)
            for i in range(len(val)):
                predict_days[val[i]] = predict[i]
        predict_days = np.array(predict_days)
        days = m.inverse_transform(days)
        with open(self.predict_result_path + self.pro_name + '.txt', 'w+') as file:
            for i in range(len(predict_days)):
                file.write(str(predict_days[i]) + '\n')
        return predict_days, days

    def variance(self, original, predict):
        return explained_variance_score(original, predict)

    def mae(self, original, predict):
        return mean_absolute_error(original, predict)

    def mse(self, original, predict):
        return mean_squared_error(original, predict)

    def cae(self, original, predict):
        return median_absolute_error(original, predict)

    def rs(self, original, predict):
        return r2_score(original, predict)

    def mape(self, original, predict):
        return mean_absolute_percentage_error(original,predict)

    def getMAESection(self, original, predict, mae):
        section0_2, section2_4, section4_6, section6_8, section8_10, section10_12, section12_14, section14_ = 0, 0, 0, 0, 0, 0, 0, 0
        gt, st = 0, 0
        for i in range(len(original)):
            ae = np.absolute(predict[i] - original[i])
            if (ae >= 0) and (ae < 2):
                section0_2 += 1
            elif (ae >= 2) and (ae < 4):
                section2_4 += 1
            elif (ae >= 4) and (ae < 6):
                section4_6 += 1
            elif (ae >= 6) and (ae < 8):
                section6_8 += 1
            elif (ae >= 8) and (ae < 10):
                section8_10 += 1
            elif (ae >= 10) and (ae < 12):
                section10_12 += 1
            elif (ae >= 12) and (ae < 14):
                section12_14 += 1
            else:
                section14_ += 1
            if ae > mae:
                gt += 1
            else:
                st += 1
        print('[0~2]: {}\t[2~4]: {}\t[4~6]: {}\t[6~8]: {}\t[8~10]: {}\t[10~12]: {}\t[12~14]: {}\t[14+]: {}'.format(
            section0_2, section2_4,
            section4_6, section6_8,
            section8_10, section10_12, section12_14, section14_))
        print('[greater_than_mae]: {}\t[smaller_than_mae]: {}'.format(gt, st))

    def getEvaluationResults(self, i):
        predict, original = self.predict(i)
        original, predict = original.squeeze(), predict.squeeze()
        var = self.variance(original, predict)
        mae = self.mae(original, predict)
        mse = self.mse(original, predict)
        cae = self.cae(original, predict)
        rs = self.rs(original, predict)
        mape = self.mape(original, predict)
        print(
            '[mean_absolute_error]: %.2f\t[mean_squared_error]: %.2f\t[mean_absolute_percentage_error]: %.2f\t[variance]: %.2f\t[median_absolute_error]: %.2f\t[r2_score]: %.2f\t'
            % (mae, mse, mape, var, cae, rs))
        self.getMAESection(original, predict, mae)

    def getDetailInformation(self):
        data = json.load(open(self.detail_clinical_path + self.pro_name + '.json'))
        return data

    def print_detail(self, i):
        predict_days, days = self.predict(i)
        data = self.getDetailInformation()
        pointer = 0
        print('submitter_id/ae/days_to_death/age/gender/weight/height/bmi/treatments/stages')
        for item in data:
            if ("days_to_death" in item.keys()) and (item["days_to_death"] <= 1825) and (item["days_to_death"] > 0):
                print(item["submitter_id"], '/', np.abs(float(days[pointer]) - float(predict_days[pointer])), '/',
                      item["days_to_death"], '/', item["age"], '/', item["gender"], '/', item["weight"], '/',
                      item["height"], '/', item["bmi"], '/', item["treatment_list"], '/', item["stage_list"], '/')
                pointer += 1

