import copy
import gc
import time
import warnings
from time import sleep

import numpy as np
from sklearn.feature_selection import SelectFromModel
from sklearn.linear_model import LinearRegression
from sklearn.svm import LinearSVR

from utils.data_process import DataProcessing

warnings.filterwarnings('ignore')


class FeatureSelection:
    def __init__(self, pro_name='COAD', model_name='VAE'):
        self.pro_name = pro_name
        self.model_name = model_name
        self.fs_data_path = 'dataset/fs_data/'
        self.fc_data_path = 'dataset/fc_data/'
        self.fs_header_path = 'dataset/fs_data/headers/'
        self.fc_header_path = 'dataset/fc_data/headers/'
        self.dp = DataProcessing(pro_name=pro_name, model_name=model_name)

    def get_data_after_lsvr(self, k, matrix, days):
        if k == 0:
            k = 1
        model = LinearSVR()
        model.fit(matrix, days)
        weights = np.abs(model.coef_)
        mask = np.argsort(-weights)[:k]
        matrix = matrix[:, mask]
        return matrix, days, mask

    def get_data_after_lr(self, k, matrix, days, mask):
        if k == 0:
            k = 1
        mask = np.array([mask])
        model = LinearRegression()
        model.fit(matrix, days)
        sfm = SelectFromModel(estimator=model, threshold=-np.inf, max_features=k, prefit=True)
        matrix = sfm.transform(matrix)
        mask = sfm.transform(mask)
        mask = mask.squeeze()
        return matrix, days, mask

    def get_k1k2_result(self, matrix, days, m, k1_list, k2_list, n_splits=10):
        result_list = []
        for k1 in k1_list:
            if k1 <= matrix.shape[1]:
                matrix_x, days_x, mask_x = self.get_data_after_lsvr(k1, matrix, days)
                for k2 in k2_list:
                    if k1 > k2:
                        matrix_y, days_y, mask_y = self.get_data_after_lr(k2, matrix_x, days_x, mask_x)
                        mae = self.dp.cross_validation(matrix_y, days_y, m, n_splits)
                        result_list.append([k1, k2, mae])
                        del matrix_y, days_y, mask_y
                        gc.collect()
                        sleep(2)
                del matrix_x, days_x, mask_x
                gc.collect()
                sleep(2)
        if len(result_list) == 0:
            result_list.append([0, 0, 1000000])
        result_list = sorted(result_list, key=lambda x: x[2])
        return result_list[0][0], result_list[0][1], result_list[0][2]

    def get_data_after_backFS(self, matrix, days, mask, m, n_splits=10):
        best_val_mae = []
        point_list = []
        matrix_t = copy.deepcopy(matrix)
        while matrix_t.shape[1] != 1:
            mae_list = np.zeros(matrix_t.shape[1])
            for time in range(matrix_t.shape[1]):
                x = matrix_t
                x = np.delete(x, obj=time, axis=1)
                val_mae = self.dp.cross_validation(x, days, m, n_splits)
                mae_list[time] = val_mae
            best_val_mae.append(mae_list.min())
            point = np.nanargmin(mae_list)
            point_list.append(point)
            matrix_t = np.delete(matrix_t, obj=point, axis=1)
        best_val_mae = np.array(best_val_mae)
        matrix_len = np.nanargmin(best_val_mae)
        for i in range(matrix_len):
            matrix = np.delete(matrix, obj=point_list[i], axis=1)
            mask = np.delete(mask, obj=point_list[i])
        return matrix, mask, best_val_mae.min()

    def get_first_time(self):
        k1_list = np.arange(20, 1200, 50)
        k2_list = np.arange(10, 1000, 20)
        matrix = self.dp.get_regression_matrix()
        days, m = self.dp.get_days_standardScaler()
        k1, k2, val_mae = self.get_k1k2_result(matrix, days, m, k1_list, k2_list, n_splits=10)
        matrix, days, mask = self.get_data_after_lsvr(k1, matrix, days)
        matrix, days, mask = self.get_data_after_lr(k2, matrix, days, mask)
        matrix, mask, val_mae = self.get_data_after_backFS(matrix, days, mask, m, n_splits=10)
        with open(self.fs_header_path + self.pro_name + '.txt', 'w+') as header_file:
            for item in mask:
                header_file.write(str(item) + '\t')
        with open(self.fc_data_path + self.pro_name + '.txt', 'w+')as file:
            for k in range(matrix.shape[0]):
                for j in range(matrix.shape[1]):
                    file.write(str(matrix[k][j]) + '\t')
                file.write('\n')
        fea_num = matrix.shape[1]
        val_mae = self.dp.cross_validation(matrix, days, m, n_splits=10)
        del matrix, days, mask
        gc.collect()
        time.sleep(2)
        print('[k1]: {}\t[k2]: {}\t[k3]: {}\t[mae]: {}'.format(k1, k2, fea_num, val_mae))

    def inter_fs(self, i, matrix, days, m, k1, k2):
        matrix, days, mask = self.get_data_after_lsvr(k1, matrix, days)
        matrix, days, mask = self.get_data_after_lr(k2, matrix, days, mask)
        matrix, mask, mae = self.get_data_after_backFS(matrix, days, mask, m, n_splits=10)
        with open(self.fc_header_path + self.model_name + '_' + self.pro_name + '_' + str(i) + '.txt',
                  'w+') as header_file:
            for item in mask:
                header_file.write(str(item) + '\t')
        with open(self.fc_data_path + self.model_name + '_' + self.pro_name + '_' + str(i) + '.txt', 'w+') as file:
            for k in range(matrix.shape[0]):
                for j in range(matrix.shape[1]):
                    file.write(str(matrix[k][j]) + '\t')
                file.write('\n')
        mae = self.dp.cross_validation(matrix, days, m, n_splits=10)
        print('[mae]: %.6f\t[FNum]: %3d\t' % (mae, matrix.shape[1]))
        return mae, matrix.shape[1]
