import warnings

import numpy as np
import torch
import torch.nn as nn

from utils.data_process import DataProcessing

warnings.filterwarnings('ignore')
np.set_printoptions(precision=3)


class S_AEWGAN:
    def __init__(self, epochs=50, batch_size=10, lr=1e-4, n1=400, n2=400, gradient=200, num_heads=10, pro_name='COAD'):
        self.dp = DataProcessing(pro_name)
        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr
        self.n1, self.n2, self.gradient, self.num_heads = n1, n2, gradient, num_heads
        self.pro_name = pro_name
        self.loss_func = nn.HingeEmbeddingLoss()

    def sparse_loss(self, input, hidden, output):
        def kl(p, q):
            q = torch.nn.functional.softmax(q, dim=0)
            q = torch.sum(q,
                          dim=0) / self.batch_size
            s1 = torch.sum(p * torch.log(p / q))
            return s1

        criterion = nn.MSELoss()
        MSE_loss = criterion(input, output)
        rho_hat = torch.sum(hidden, dim=0, keepdim=True)
        sparsity_penalty = 3 * kl(0.01, rho_hat)
        return MSE_loss + sparsity_penalty

    def pre_train(self, dataLoader, generator, g_optimizer):
        for index, (order, label, data) in enumerate(dataLoader):
            generator.train()
            out1, out2, out3 = generator(data)
            loss = self.sparse_loss(data, out2, out3)
            g_optimizer.zero_grad()
            loss.backward()
            g_optimizer.step()
        return generator, g_optimizer

    def train(self, dataLoader, data_size, generator, g_optimizer, discriminator, d_optimizer):
        train_loss = 0
        hid1, hid2, out, order_list = [], [], [], []
        for index, (order, label, data) in enumerate(dataLoader):
            generator.train()
            discriminator.train()
            for parm in discriminator.parameters():
                parm.data.clamp_(-1, 1)
            real_out = discriminator(data)
            real_out = real_out.view([len(data)])
            _, _, fake_data = generator(data)
            fake_out = discriminator(fake_data)
            fake_out = fake_out.view([len(data)])
            d_loss = -torch.mean(real_out) + torch.mean(fake_out)
            d_optimizer.zero_grad()
            d_loss.backward()
            d_optimizer.step()

            out1, out2, out3 = generator(data)
            output = discriminator(out3)
            train_loss += np.abs(d_loss.item() / 2) * data.shape[0]
            g_loss = self.sparse_loss(data, output, out3)
            g_optimizer.zero_grad()
            g_loss.backward()
            g_optimizer.step()
            hid1.append(out1)
            hid2.append(out2)
            out.append(out3)
            order_list.append(order)
        return train_loss / data_size, generator, g_optimizer, discriminator, d_optimizer, hid1, hid2, out, order_list

    def test(self, dataLoader, data_size, generator, discriminator):
        test_loss = 0
        hid1, hid2, out, order_list = [], [], [], []
        for index, (order, label, data) in enumerate(dataLoader):
            generator.eval()
            out1, out2, out3 = generator(data)
            real_out = discriminator(data)
            fake_out = discriminator(out3)
            real_out = real_out.view([len(data)])
            fake_out = fake_out.view([len(data)])
            d_loss = -torch.mean(real_out) + torch.mean(fake_out)
            test_loss += np.abs(d_loss.item() / 2) * data.shape[0]
            hid1.append(out1)
            hid2.append(out2)
            out.append(out3)
            order_list.append(order)
        return test_loss / data_size, hid1, hid2, out, order_list
