import copy
import heapq
import warnings

import numpy as np
import pandas as pd
from sklearn.feature_selection import SelectFromModel
from sklearn.linear_model import LinearRegression, Lasso, Ridge
from sklearn.metrics import mean_absolute_error, mean_squared_error
from sklearn.model_selection import KFold
from sklearn.preprocessing import StandardScaler
from sklearn.svm import LinearSVR

warnings.filterwarnings('ignore')


class Compare:
    def __init__(self, pro_name='COAD', method_name='Crystall'):
        self.pro_name = pro_name
        self.method_name = method_name
        self.fs_header_path = '../dataset/fs_data/headers/'
        self.final_matrix_path = '../dataset/raw_data/final_matrix/'
        self.final_days_path = '../dataset/raw_data/final_days/'
        self.compare_matrix_path = '../dataset/fs_data/compare/'

    def evaluate(self, matrix_train, days_train, matrix, days, m, model):
        model.fit(matrix_train, days_train)
        predict = model.predict(matrix)
        predict, days = m.inverse_transform(predict), m.inverse_transform(days)
        return mean_absolute_error(predict, days), mean_squared_error(predict, days)

    def cross_validation(self, matrix, days, m, model, n_splits=10):
        lsvr_mae = []
        lsvr_mse = []
        kf = KFold(n_splits=n_splits, random_state=0, shuffle=True)
        for train, val in kf.split(matrix, days):
            t_data, t_label = np.array(matrix)[train], np.array(days)[train]
            v_data, v_label = np.array(matrix)[val], np.array(days)[val]
            mae, mse = self.evaluate(t_data, t_label, v_data, v_label, m, model)
            lsvr_mae.append(mae)
            lsvr_mse.append(mse)
        lsvr_mae, lsvr_mse = np.array(lsvr_mae), np.array(lsvr_mse)
        return lsvr_mae.mean(), lsvr_mse.mean()

    def get_data_after_backFS(self, matrix, days, mask, m, n_splits=10):
        best_val_mae = []
        best_val_mse = []
        point_list = []
        matrix_t = copy.deepcopy(matrix)
        while matrix_t.shape[1] != 1:
            mae_list = np.zeros(matrix_t.shape[1])
            mse_list = np.zeros(matrix_t.shape[1])
            for time in range(matrix_t.shape[1]):
                x = matrix_t
                x = np.delete(x, obj=time, axis=1)
                model = LinearSVR()
                val_mae, val_mse = self.cross_validation(x, days, m, model, n_splits)
                mae_list[time] = val_mae
                mse_list[time] = val_mse
            best_val_mae.append(mae_list.min())
            point = np.nanargmin(mae_list)
            point_list.append(point)
            matrix_t = np.delete(matrix_t, obj=point, axis=1)
            best_val_mse.append(mse_list[point])
            # print('[mae]: %.6f\t[FNum]: %3d\t[index]: %3d' % (best_val_mae[-1], matrix_t.shape[1], point))
        best_val_mae, best_val_mse = np.array(best_val_mae), np.array(best_val_mse)
        matrix_len = np.nanargmin(best_val_mae)
        for i in range(matrix_len):
            matrix = np.delete(matrix, obj=point_list[i], axis=1)
            mask = np.delete(mask, obj=point_list[i])
        return matrix, mask, best_val_mae.min(), best_val_mse[matrix_len]

    # Lasso
    def lasso(self, matrix, days, m):
        mask = np.arange(0, matrix.shape[1])
        mask = [mask]
        mask = np.array(mask)
        mae_list = []
        for k in range(30, 300, 10):
            model = Lasso()
            model.fit(matrix, days)
            threshold = sorted(abs(model.coef_), reverse=True)[k - 1]
            sfm = SelectFromModel(model, prefit=True, threshold=threshold)
            matrix_t = sfm.transform(matrix)
            mae, mse = self.cross_validation(matrix_t, days, m, model, n_splits=10)
            mae, mse = float('%.2f' % mae), float('%.2f' % mse)
            # print('[k]: {}\t[mae]: {}\t[mse]: {}'.format(k,mae,mse))
            mae_list.append(mae)
        k = np.nanargmin(np.array(mae_list))
        model = Lasso()
        model.fit(matrix, days)
        threshold = sorted(abs(model.coef_), reverse=True)[(k + 3) * 10]
        sfm = SelectFromModel(model, prefit=True, threshold=threshold)
        matrix = sfm.transform(matrix)
        mask = sfm.transform(mask)
        mae, mse = self.cross_validation(matrix, days, m, model, n_splits=10)
        mae, mse = float('%.2f' % mae), float('%.2f' % mse)
        return matrix, mask, mae, mse

    # Ridge
    def ridge(self, matrix, days, m):
        mask = np.arange(0, matrix.shape[1])
        mask = [mask]
        mask = np.array(mask)
        mae_list = []
        for k in range(30, 300, 10):
            model = Ridge()
            model.fit(matrix, days)
            threshold = sorted(abs(model.coef_), reverse=True)[k - 1]
            sfm = SelectFromModel(model, prefit=True, threshold=threshold)
            matrix_t = sfm.transform(matrix)
            mae, mse = self.cross_validation(matrix_t, days, m, model, n_splits=10)
            mae, mse = float('%.2f' % mae), float('%.2f' % mse)
            # print('[k]: {}\t[mae]: {}\t[mse]: {}'.format(k, mae, mse))
            mae_list.append(mae)
        k = np.nanargmin(np.array(mae_list))
        model = Ridge()
        model.fit(matrix, days)
        threshold = sorted(abs(model.coef_), reverse=True)[(k + 3) * 10]
        sfm = SelectFromModel(model, prefit=True, threshold=threshold)
        matrix = sfm.transform(matrix)
        mask = sfm.transform(mask)
        mae, mse = self.cross_validation(matrix, days, m, model, n_splits=10)
        mae, mse = float('%.2f' % mae), float('%.2f' % mse)
        return matrix, mask, mae, mse

    # LinearSVR
    def lsvr(self, matrix, days, m):
        mask = np.arange(0, matrix.shape[1])
        mask = [mask]
        mask = np.array(mask)
        mae_list = []
        for k in range(30, 300, 10):
            model = LinearSVR()
            model.fit(matrix, days)
            threshold = sorted(abs(model.coef_), reverse=True)[k - 1]
            sfm = SelectFromModel(model, prefit=True, threshold=threshold)
            matrix_t = sfm.transform(matrix)
            mae, mse = self.cross_validation(matrix_t, days, m, model, n_splits=10)
            mae, mse = float('%.2f' % mae), float('%.2f' % mse)
            # print('[k]: {}\t[mae]: {}\t[mse]: {}'.format(k, mae, mse))
            mae_list.append(mae)
        k = np.nanargmin(np.array(mae_list))
        model = LinearSVR()
        model.fit(matrix, days)
        threshold = sorted(abs(model.coef_), reverse=True)[(k + 3) * 10]
        sfm = SelectFromModel(model, prefit=True, threshold=threshold)
        matrix = sfm.transform(matrix)
        mask = sfm.transform(mask)
        mae, mse = self.cross_validation(matrix, days, m, model, n_splits=10)
        mae, mse = float('%.2f' % mae), float('%.2f' % mse)
        return matrix, mask, mae, mse

    # LinearRegression
    def lr(self, matrix, days, m):
        mask = np.arange(0, matrix.shape[1])
        mask = [mask]
        mask = np.array(mask)
        mae_list = []
        for k in range(30, 300, 10):
            model = LinearRegression()
            model.fit(matrix, days)
            threshold = sorted(abs(model.coef_), reverse=True)[k - 1]
            sfm = SelectFromModel(model, prefit=True, threshold=threshold)
            matrix_t = sfm.transform(matrix)
            mae, mse = self.cross_validation(matrix_t, days, m, model, n_splits=10)
            mae, mse = float('%.2f' % mae), float('%.2f' % mse)
            # print('[k]: {}\t[mae]: {}\t[mse]: {}'.format(k, mae, mse))
            mae_list.append(mae)
        k = np.nanargmin(np.array(mae_list))
        model = LinearRegression()
        model.fit(matrix, days)
        threshold = sorted(abs(model.coef_), reverse=True)[(k + 3) * 10]
        sfm = SelectFromModel(model, prefit=True, threshold=threshold)
        matrix = sfm.transform(matrix)
        mask = sfm.transform(mask)
        mae, mse = self.cross_validation(matrix, days, m, model, n_splits=10)
        mae, mse = float('%.2f' % mae), float('%.2f' % mse)
        return matrix, mask, mae, mse

    # crystall
    def crystall(self, matrix, days, m):
        mask = np.arange(0, matrix.shape[1])
        mask = [mask]
        mask = np.array(mask)
        lsvr = LinearSVR()
        lsvr.fit(matrix, days)
        k_list = [50, 100, 150, 200, 250, 300, 350, 400]
        mae_list = []
        for k in k_list:
            threshold = heapq.nlargest(k, abs(lsvr.coef_))[-1]
            sfm = SelectFromModel(estimator=lsvr, threshold=threshold, prefit=True)
            matrix_t = sfm.transform(matrix)
            model = LinearSVR()
            mae, mse = self.cross_validation(matrix_t, days, m, model=model, n_splits=10)
            mae_list.append(mae)
        mae_list = np.array(mae_list)
        k = k_list[mae_list.argmin()]
        sfm = SelectFromModel(estimator=lsvr, threshold=-np.inf, max_features=k, prefit=True)
        matrix = sfm.transform(matrix)
        mask = sfm.transform(mask)
        matrix, mask, mae, mse = self.get_data_after_backFS(matrix, days, mask, m, n_splits=10)
        model = LinearSVR()
        mae, mse = self.cross_validation(matrix, days, m, model, n_splits=10)
        return matrix, mask, mae, mse

    def get_results(self):
        matrix = pd.read_table(self.final_matrix_path + self.pro_name + '.txt', sep='\t', header=None,
                               low_memory=False).values
        matrix = matrix[:, :-1]
        days = pd.read_table(self.final_days_path + self.pro_name + '.txt', header=None).values
        m = StandardScaler()
        days = m.fit_transform(days)
        days = days.squeeze()
        func_dict = {'Lasso': 'lasso', 'Ridge': 'ridge', 'LinearRegression': 'lr', 'LinearSVR': 'lsvr',
                     'Crystall': 'crystall'}
        fs_func = getattr(self, func_dict[self.method_name])
        matrix, mask, mae, mse = fs_func(matrix, days, m)
        print('[method_name]: {}\t[mae]: {}\t[mse]: {}'.format(self.method_name, mae, mse))
        with open(self.compare_matrix_path + self.method_name + '_' + self.pro_name + '.txt', 'w+') as mf:
            for i in range(matrix.shape[0]):
                for j in range(matrix.shape[1]):
                    mf.write(str(matrix[i][j]) + '\t')
                mf.write('\n')
        with open(self.fs_header_path + self.method_name + '_' + self.pro_name + '.txt', 'w+') as hf:
            for item in mask:
                hf.write(str(item) + '\t')


# pro_list = ['BRCA', 'COAD', 'BLCA', 'UCEC', 'LGG', 'CESC', 'SARC', 'STAD', 'LIHC', 'LUAD']
# method_list = ['Crystall', 'Lasso', 'Ridge', 'LinearRegression', 'LinearSVR']
# for pro_name in pro_list:
#     print("======" + pro_name + "======")
#     for method in method_list:
#         c = Compare(pro_name, method)
#         c.get_results()
#
# c = Compare('COAD', 'LinearSVR')
# c.get_results()
